% P4P + unknown focal length helper
% given a set of 4x 2D<->3D correspondences, calculate focal length and 3 unknown depths
%
% by Martin Bujnak, (c)apr2008
%
%
% Please refer to the following paper, when using this code : 
%
%      Bujnak, M., Kukelova, Z., and Pajdla, T. A general solution to the p4p
%      problem for camera with unknown focal length. CVPR 2008, Anchorage, 
%      Alaska, USA, June 2008
%
% [f zb zc zd] = p4pfcode(glab, glac, glad, glbc, glbd, glcd, a1, a2, b1, b2, c1, c2, d1, d2)
% 
% input:
%   glXY - ||X-Y||^2 - quadratic distances between 3D points X and Y
%   a1 (a2) = x (resp y) measurement of the first 2D point
%   b1 (b2) = x (resp y) measurement of the second 2D point
%   c1 (c2) = x (resp y) measurement of the third 2D point
%   d1 (d2) = x (resp y) measurement of the fourth 2D point
%
% output:
%   f - estimated focal lengths
%   zb, zc, zd  - depths of points b, c, d (depth of 'a' is fixed to 1)
% 
function [f zb zc zd] = p4pfcode(glab, glac, glad, glbc, glbd, glcd, a1, a2, b1, b2, c1, c2, d1, d2)

	% precalculate polynomial equations coefficients
	c(1) = 1;
	c(2) = 1/2/glad*glbc-1/2*glab/glad-1/2*glac/glad;
	c(3) = -1;
	c(4) = -1;
	c(5) = c2*b2+c1*b1;
	c(6) = glac/glad-1/glad*glbc+glab/glad;
	c(7) = 1/2/glad*glbc*d2^2-1/2*glab/glad*d2^2-1/2*glac/glad*d2^2-1/2*glac/glad*d1^2+1/2/glad*glbc*d1^2-1/2*glab/glad*d1^2;
	c(8) = 1-1/2*glac/glad-1/2*glab/glad+1/2/glad*glbc;
	c(9) = -b1*a1-a2*b2;
	c(10) = -c2*a2-c1*a1;
	c(11) = -a1/glad*glbc*d1+a1*glac/glad*d1+glac/glad*a2*d2+a1*glab/glad*d1-1/glad*glbc*a2*d2+glab/glad*a2*d2;
	c(12) = a2^2+a1^2-1/2*glac/glad*a2^2-1/2*a1^2*glac/glad+1/2/glad*glbc*a2^2-1/2*a1^2*glab/glad+1/2*a1^2/glad*glbc-1/2*glab/glad*a2^2;
	c(13) = 1;
	c(14) = -glac/glad;
	c(15) = -2;
	c(16) = c1^2+c2^2;
	c(17) = 2*glac/glad;
	c(18) = -glac/glad*d1^2-glac/glad*d2^2;
	c(19) = -glac/glad+1;
	c(20) = -2*c2*a2-2*c1*a1;
	c(21) = 2*a1*glac/glad*d1+2*glac/glad*a2*d2;
	c(22) = -glac/glad*a2^2+a2^2+a1^2-a1^2*glac/glad;
	c(23) = 1;
	c(24) = 1/2/glad*glbd-1/2-1/2*glab/glad;
	c(25) = -1;
	c(26) = glab/glad-1/glad*glbd;
	c(27) = d2*b2+b1*d1;
	c(28) = -1/2*glab/glad*d2^2-1/2*glab/glad*d1^2+1/2/glad*glbd*d2^2+1/2/glad*glbd*d1^2-1/2*d2^2-1/2*d1^2;
	c(29) = -1/2*glab/glad+1/2/glad*glbd+1/2;
	c(30) = -a2*b2-b1*a1;
	c(31) = -a1/glad*glbd*d1+a1*glab/glad*d1+glab/glad*a2*d2-1/glad*glbd*a2*d2;
	c(32) = 1/2/glad*glbd*a2^2+1/2*a1^2/glad*glbd-1/2*glab/glad*a2^2-1/2*a1^2*glab/glad+1/2*a1^2+1/2*a2^2;
	c(33) = 1;
	c(34) = -1/2*glac/glad+1/2*glcd/glad-1/2;
	c(35) = -1;
	c(36) = glac/glad-glcd/glad;
	c(37) = c1*d1+c2*d2;
	c(38) = 1/2*glcd/glad*d1^2-1/2*glac/glad*d2^2-1/2*glac/glad*d1^2-1/2*d1^2-1/2*d2^2+1/2*glcd/glad*d2^2;
	c(39) = -1/2*glac/glad+1/2+1/2*glcd/glad;
	c(40) = -c1*a1-c2*a2;
	c(41) = glac/glad*a2*d2+a1*glac/glad*d1-glcd/glad*a2*d2-glcd*a1/glad*d1;
	c(42) = 1/2*a1^2+1/2*a2^2-1/2*glac/glad*a2^2+1/2*glcd*a1^2/glad-1/2*a1^2*glac/glad+1/2*glcd/glad*a2^2;

	M = zeros(78, 88);
	ci = [72, 149, 520, 597, 752, 829, 1062, 1217, 1528, 1895, 2050, 2127, 2360, 2515, 2904, 3439, 3594, 3983, 4993];
	M(ci) = c(1);

	ci = [384, 461, 988, 1299, 1454, 1609, 1686, 1841, 2830, 2985, 3140, 3295, 3372, 4219, 4374, 4607, 5539];
	M(ci) = c(2);

	ci = [618, 929, 1924, 2235, 2390, 2545, 2778, 2933, 3244, 3454, 3609, 3842, 3997, 4308, 4843, 4998, 5231, 5851];
	M(ci) = c(3);

	ci = [696, 1007, 2002, 2313, 2468, 2623, 2856, 3011, 3322, 3532, 3687, 3920, 4075, 4386, 4921, 5076, 5309, 5929];
	M(ci) = c(4);

	ci = [774, 1085, 2080, 2391, 2546, 2701, 2934, 3089, 3455, 3610, 3765, 3998, 4153, 4464, 4999, 5154, 5387, 6007];
	M(ci) = c(5);

	ci = [1008, 1319, 2314, 2858, 3013, 3168, 3323, 3400, 3922, 4077, 4232, 4387, 4620, 5310, 5543, 6163];
	M(ci) = c(6);

	ci = [1476, 1709, 2860, 3171, 3326, 3402, 4235, 4390, 4545, 4622, 4777, 5545, 5700, 5777, 6397];
	M(ci) = c(7);

	ci = [2334, 3950, 4105, 4260, 4415, 4648, 4936, 5091, 5323, 5556, 5934, 6167, 6475];
	M(ci) = c(8);

	ci = [2412, 2801, 3484, 3873, 4028, 4183, 4338, 4493, 4726, 4859, 5014, 5169, 5246, 5401, 5634, 5857, 6012, 6245, 6553];
	M(ci) = c(9);

	ci = [2490, 2879, 3562, 3951, 4106, 4416, 4571, 4804, 4937, 5092, 5324, 5479, 5712, 5935, 6090, 6323, 6631];
	M(ci) = c(10);

	ci = [2880, 3191, 3952, 4263, 4418, 4573, 4650, 4805, 5326, 5481, 5558, 5713, 5790, 6169, 6324, 6401, 6709];
	M(ci) = c(11);

	ci = [3972, 4283, 4966, 5354, 5509, 5586, 5741, 5818, 5950, 6105, 6182, 6337, 6414, 6481, 6636, 6713, 6787];
	M(ci) = c(12);

	ci = [74, 229, 527, 682, 759, 836, 1147, 1224, 1613, 1980, 2057, 2134, 2445, 2522, 2599, 2988, 3520, 3597, 4064, 5072];
	M(ci) = c(13);

	ci = [308, 463, 917, 1306, 1383, 1538, 1693, 1770, 2759, 2914, 3147, 3224, 3301, 3378, 4222, 4299, 4610, 5540];
	M(ci) = c(14);

	ci = [620, 1009, 1931, 2320, 2397, 2552, 2863, 2940, 3329, 3461, 3616, 3927, 4004, 4081, 4392, 4924, 5001, 5312, 5930];
	M(ci) = c(15);

	ci = [776, 1165, 2087, 2476, 2553, 2708, 3019, 3096, 3540, 3617, 3772, 4083, 4160, 4548, 5080, 5157, 5468, 6086];
	M(ci) = c(16);

	ci = [932, 1321, 2243, 2787, 2942, 3175, 3252, 3407, 3851, 4006, 4239, 4316, 4393, 4626, 5235, 5546, 6164];
	M(ci) = c(17);

	ci = [1400, 1711, 2789, 3178, 3255, 3409, 4242, 4319, 4474, 4629, 4706, 4783, 5548, 5625, 5780, 6398];
	M(ci) = c(18);

	ci = [2258, 3879, 4034, 4267, 4344, 4655, 4865, 5020, 5252, 5329, 5562, 5859, 6170, 6476];
	M(ci) = c(19);

	ci = [2414, 2881, 3491, 3958, 4035, 4190, 4423, 4500, 4811, 4944, 5021, 5176, 5331, 5408, 5485, 5718, 5938, 6015, 6326, 6632];
	M(ci) = c(20);

	ci = [2804, 3193, 3881, 4270, 4347, 4502, 4657, 4734, 5255, 5410, 5565, 5642, 5719, 5796, 6172, 6249, 6404, 6710];
	M(ci) = c(21);

	ci = [3896, 4285, 4895, 5283, 5438, 5593, 5670, 5825, 5879, 6034, 6189, 6266, 6343, 6420, 6484, 6561, 6716, 6788];
	M(ci) = c(22);

	ci = [154, 309, 609, 920, 1075, 1386, 2220, 2375, 2530, 2763, 2918, 3229, 3835, 3990, 4301, 5229];
	M(ci) = c(23);

	ci = [388, 465, 999, 1310, 1465, 1698, 2843, 2998, 3153, 3308, 3385, 4225, 4380, 4613, 5541];
	M(ci) = c(24);

	ci = [622, 933, 1935, 2246, 2401, 2790, 3467, 3622, 3855, 4010, 4321, 4849, 5004, 5237, 5853];
	M(ci) = c(25);

	ci = [1012, 1323, 2325, 2869, 3180, 3935, 4090, 4245, 4400, 4633, 5316, 5549, 6165];
	M(ci) = c(26);

	ci = [1090, 1401, 2403, 2792, 2947, 3258, 3858, 4013, 4168, 4323, 4478, 4711, 5239, 5394, 5627, 6243];
	M(ci) = c(27);

	ci = [1480, 1713, 2871, 3182, 3337, 3414, 4248, 4403, 4558, 4635, 4790, 5551, 5706, 5783, 6399];
	M(ci) = c(28);

	ci = [2338, 3961, 4272, 4949, 5104, 5336, 5569, 5940, 6173, 6477];
	M(ci) = c(29);

	ci = [2416, 2805, 3495, 3884, 4039, 4350, 4872, 5027, 5182, 5259, 5414, 5647, 5863, 6018, 6251, 6555];
	M(ci) = c(30);

	ci = [2884, 3195, 3963, 4274, 4429, 4662, 5339, 5494, 5571, 5726, 5803, 6175, 6330, 6407, 6711];
	M(ci) = c(31);

	ci = [3976, 4287, 4977, 5365, 5598, 5963, 6118, 6195, 6350, 6427, 6487, 6642, 6719, 6789];
	M(ci) = c(32);

	ci = [234, 389, 694, 849, 1004, 1159, 1470, 1547, 1624, 2307, 2384, 2461, 2538, 2615, 2848, 2925, 3002, 3313, 3917, 3994, 4071, 4382, 5308];
	M(ci) = c(33);

	ci = [390, 467, 1006, 1239, 1316, 1471, 1704, 1781, 1858, 2774, 2851, 2928, 3005, 3160, 3237, 3314, 3391, 4229, 4306, 4383, 4616, 5542];
	M(ci) = c(34);

	ci = [702, 1013, 2020, 2175, 2330, 2485, 2874, 2951, 3028, 3476, 3553, 3630, 3707, 3940, 4017, 4094, 4405, 4931, 5008, 5085, 5318, 5932];
	M(ci) = c(35);

	ci = [1014, 1325, 2332, 2565, 2875, 3186, 3263, 3340, 3866, 3943, 4020, 4097, 4252, 4329, 4406, 4639, 5242, 5319, 5552, 6166];
	M(ci) = c(36);

	ci = [1170, 1481, 2488, 2721, 2876, 3031, 3342, 3945, 4022, 4099, 4176, 4408, 4485, 4562, 4795, 5321, 5398, 5475, 5708, 6322];
	M(ci) = c(37);

	ci = [1482, 1715, 2878, 3111, 3188, 3343, 3420, 4257, 4334, 4411, 4488, 4565, 4642, 4719, 4796, 5555, 5632, 5709, 5786, 6400];
	M(ci) = c(38);

	ci = [2340, 3657, 3967, 4278, 4355, 4432, 4880, 4957, 5034, 5111, 5265, 5342, 5575, 5866, 5943, 6176, 6478];
	M(ci) = c(39);

	ci = [2496, 2885, 3580, 3813, 3968, 4123, 4434, 4511, 4588, 4959, 5036, 5113, 5190, 5344, 5421, 5498, 5731, 5945, 6022, 6099, 6332, 6634];
	M(ci) = c(40);

	ci = [2886, 3197, 3970, 4203, 4280, 4435, 4668, 4745, 4822, 5270, 5347, 5424, 5501, 5578, 5655, 5732, 5809, 6179, 6256, 6333, 6410, 6712];
	M(ci) = c(41);

	ci = [3978, 4289, 4984, 5217, 5371, 5604, 5681, 5758, 5894, 5971, 6048, 6125, 6202, 6279, 6356, 6433, 6491, 6568, 6645, 6722, 6790];
	M(ci) = c(42);


	Mr = rref(M);

	A = zeros(10);
	amcols = [88 87 86 85 84 83 82 81 80 79];
	A(1, 2) = 1;
	A(2, 6) = 1;
	A(3, 7) = 1;
	A(4, 8) = 1;
	A(5, 9) = 1;
	A(6, :) = -Mr(75, amcols);
	A(7, :) = -Mr(74, amcols);
	A(8, :) = -Mr(73, amcols);
	A(9, :) = -Mr(72, amcols);
	A(10, :) = -Mr(71, amcols);

	[V D] = eig(A);
	sol =  V([2, 3, 4, 5],:)./(ones(4,1)*V(1,:));

	if (find(isnan(sol(:))) > 0)
		
		f = [];
		zb = [];
		zc = [];
		zd = [];
	else
		
		I = find(not(imag( sol(4,:) )));
        fidx = find(sol(4, I) > 0);
    
        f = sqrt(sol(4, I(fidx)));
		zd = sol(1, I(fidx));
		zc = sol(2, I(fidx));
		zb = sol(3, I(fidx));
	end
end
